"""Resize images on the website to be only as large as they need to be.

Takes a file containing the filesystem paths to images generated by running
find from withe httpdocs directory on the server:

`find file -iname '*.jpg' -print0 > ../private/temp/file-jpg`

TODO: Rewrite as a Django management command.

"""
# Hack to run it as if inside parent dir
import sys
from os.path import dirname, abspath
sys.path[0] = dirname(dirname(abspath(__file__)))

from django.core.management import setup_environ
import settings
setup_environ(settings)

from urllib.parse import quote
from urllib.parse import urljoin
from vanilla.models import Item
from django.db.models import Q
import html5lib
import os
import Image

IMAGES_BASE = 'httpdocs/'
IMAGES_LIST = 'temp/file-jpg'
URL_PREFIX = ''
NS = {'x': 'http://www.w3.org/1999/xhtml'}

def item_all_html(item):
    html_blocks = [item.content, item.sub_content_b]
    html = ''.join([c for c in html_blocks if c is not None])
    return html.replace('\x16', '')

def parse_css_for_width_height(style):
    """Quick hack to parse image dimensions from inline CSS."""
    width, height = None, None
    styles = {}
    selectors = [sel.split(':') for sel in style.split(';') if sel != '']
    selectors = [s for s in selectors if len(s) == 2]
    for prop, value in selectors:
        styles[prop.strip()] = value.strip()
    if 'width' in styles and styles['width'].endswith('px'):
        try:
            width = int(styles['width'][:-2])
        except ValueError:
            pass
    if 'height' in styles and styles['height'].endswith('px'):
        try:
            height = int(styles['height'][:-2])
        except ValueError:
            pass
    return (width, height)

def max_size_in_html(img_url, html):
    """
    Find the largest dimensions for a particular image.

    Img could be used multiple times in the html block.

    Return None if not found or no dimensions.
    """
    max_width = None
    max_height = None
    tree = html5lib.parse(html, treebuilder='lxml')
    tags = tree.xpath("//x:img[@src='%s']" % img_url, namespaces=NS)
    if len(tags) == 0:
        return None
    for tag in tags:
        width = None
        height = None
        try:
            width, height = (int(tag.attrib['width']),
                             int(tag.attrib['height']))
        except (ValueError, KeyError):
            try:
                width, height = parse_css_for_width_height(tag.attrib['style'])
            except KeyError:
                pass
        if width is None and height is None:
            print("   - Found in item #%d with no dimensions, must be biggest." % item.item_id)
            return None
        elif width > max_width or height > max_height:
            print("   - Found in item #%d (%s x %s)" % (item.item_id, width, height))
            max_width = width
            max_height = height

    return (width, height)

def resize_img_file(img_path, width, height):
    if width and height:
        print("  Max size in HTML is %s x %s" % (width, height))
        # if file dimensions on disk > html dimensions, backup image and
        # resize
        try:
            file = Image.open(img_path)
        except IOError:
            print("  - Error openning file")
            return
        print("  Image on disk is %s" % str(file.size))
        if file.size[0] > width and file.size[1] > height:
            input_text = input("Continue? (y/N): ")
            if input_text.lower() == 'y':
                print("  Resizing")
                file.thumbnail((width, height), Image.BICUBIC)
                file.save(img_path, quality=90)
        else:
            print("  Image on disk correctly sized, skipping resize")
    else:
        print("  No size, skipping resize")

if __name__ == '__main__':
    imgs = open(IMAGES_LIST).read().split("\0")

    for img in [i for i in imgs if i != '']:
        img_path = os.path.join(IMAGES_BASE, img)
        print(img_path)
        url = urljoin(URL_PREFIX, quote(img, '()/'))
        print("  " + url)
        items = Item.objects.filter(Q(content__contains=url) |
                                    Q(sub_content_b__contains=url))
        max_width = None
        max_height = None
        if len(items) == 0:
            print("  Image not found in any page.")
            continue
        else:
            for item in items:
                size = max_size_in_html(url, item_all_html(item))
                if size is None:
                    # Image with no size found, so much be largest
                    max_width = None
                    max_height = None
                    break
                else:
                    width, height = size
                    if width > max_width or height > max_height:
                        max_width = width
                        max_height = height
        resize_img_file(img_path, max_width, max_height)
